import numpy as np
import tensorflow as tf
from collections import deque
import random

H1 = 128
H2 = 128
H3 = 256
BATCH_SIZE = 512
GAMMA = 0.99
INITIAL_EPSILON = 1
DECAY_RATE = 1
REPLAY_SIZE = 100000

np.random.seed(1)
tf.set_random_seed(1)
random.seed(1)

class DQN():
    def __init__(self, IsTrain=False, IsBased=False, scope=None, sess=None):
        self.replay_buffer = deque()
        self.time_step = 0
        self.epsilon = INITIAL_EPSILON
        self.state_dim = 6
        self.action_dim = 3
        self.scope = scope

        if self.scope is None:
            self.create_network_without_scope()
        else:
            self.create_network()
        self.create_training_method()
        self.IsTrain = IsTrain
        self.IsBased = IsBased
        self.saver = tf.train.Saver()
        self.session = sess
        # self.session.run(tf.global_variables_initializer())

        self.bool_defaule_action = False
        
        
    def load_parms(self):
        if self.IsTrain:
            if self.IsBased:
                self.epsilon = INITIAL_EPSILON * 0.5
                if self.scope is None:
                    checkpoint = tf.train.get_checkpoint_state("saved_networks")
                else:
                    checkpoint = tf.train.get_checkpoint_state(self.scope + "_saved_networks")
                if checkpoint and checkpoint.model_checkpoint_path:
                    self.saver.restore(self.session, checkpoint.model_checkpoint_path)
                    print("Successfully loaded:", checkpoint.model_checkpoint_path)
                else:
                    print("Could not find old network weights")
        else:
            if self.scope is None:
                checkpoint = tf.train.get_checkpoint_state("saved_networks")
            else:
                checkpoint = tf.train.get_checkpoint_state(self.scope + "_saved_networks")
            if checkpoint and checkpoint.model_checkpoint_path:
                self.saver.restore(self.session, checkpoint.model_checkpoint_path)
                print("\nSuccessfully loaded:", checkpoint.model_checkpoint_path)
            else:
                print("\n LoadNone: default action for use_agent: ")
                print("Could not find old network weights, the agent will keep choosing default_action = 1\n")
                self.bool_defaule_action = True
                
    def create_network(self):
        with tf.variable_scope(self.scope):
            self.state_input = tf.placeholder(tf.float32, [None, self.state_dim], name='state')

            self.W1 = tf.get_variable("W1", [self.state_dim, H1])
            self.b1 = tf.Variable(tf.constant(0.01, shape=[H1, ]))
            layer1 = tf.nn.relu(tf.matmul(self.state_input, self.W1) + self.b1)

            self.W2 = tf.get_variable("W2", [H1, H2])
            self.b2 = tf.Variable(tf.constant(0.01, shape=[H2, ]))
            layer2 = tf.nn.relu(tf.matmul(layer1, self.W2) + self.b2)

            self.W3 = tf.get_variable("W3", [H2, H3])
            self.b3 = tf.Variable(tf.constant(0.01, shape=[H3, ]))
            layer3 = tf.nn.relu(tf.matmul(layer2, self.W3) + self.b3)

            self.W4 = tf.get_variable("W4", [H3, self.action_dim])
            self.b4 = tf.Variable(tf.constant(0.01, shape=[self.action_dim, ]))

            self.Q_value = tf.matmul(layer3, self.W4) + self.b4

    def create_network_without_scope(self):
        self.state_input = tf.placeholder(tf.float32, [None, self.state_dim], name='state')

        self.W1 = tf.get_variable("W1", [self.state_dim, H1])
        self.b1 = tf.Variable(tf.constant(0.01, shape=[H1, ]))
        layer1 = tf.nn.relu(tf.matmul(self.state_input, self.W1) + self.b1)

        self.W2 = tf.get_variable("W2", [H1, H2])
        self.b2 = tf.Variable(tf.constant(0.01, shape=[H2, ]))
        layer2 = tf.nn.relu(tf.matmul(layer1, self.W2) + self.b2)

        self.W3 = tf.get_variable("W3", [H2, self.action_dim])
        self.b3 = tf.Variable(tf.constant(0.01, shape=[self.action_dim, ]))

        self.Q_value = tf.matmul(layer2, self.W3) + self.b3

    def create_training_method(self):
        self.action_input = tf.placeholder(tf.float32, [None, self.action_dim])
        self.y_input = tf.placeholder(tf.float32, [None])
        Q_action = tf.reduce_sum(tf.multiply(self.Q_value, self.action_input),
                                 reduction_indices=1)

        self.loss = tf.reduce_mean(tf.square(self.y_input - Q_action))
        self.optimizer = tf.train.AdamOptimizer(0.00005).minimize(self.loss)

    def perceive(self, state, action, reward, next_state, done):
        one_hot_action = np.zeros(self.action_dim)
        one_hot_action[action] = 1
        self.replay_buffer.append((state, one_hot_action, reward,
                                   next_state, done))
        if len(self.replay_buffer) > REPLAY_SIZE:
            self.replay_buffer.popleft()
        if len(self.replay_buffer) > BATCH_SIZE:
            self.train_network()

    def train_network(self):
        self.time_step += 1
        minibatch = random.sample(self.replay_buffer, BATCH_SIZE)
        state_batch = [data[0] for data in minibatch]
        action_batch = [data[1] for data in minibatch]
        reward_batch = [data[2] for data in minibatch]
        next_state_batch = [data[3] for data in minibatch]

        y_batch = []
        Q_value_batch = self.Q_value.eval(feed_dict={self.state_input: next_state_batch})
        for i in range(0, BATCH_SIZE):
            done = minibatch[i][4]
            if done:
                y_batch.append(reward_batch[i])
            else:
                y_batch.append(reward_batch[i] + GAMMA * np.max(Q_value_batch[i]))

        feed_dict = {self.y_input: y_batch,
                     self.action_input: action_batch,
                     self.state_input: state_batch}
        self.session.run(self.optimizer, feed_dict)

    def egreedy_action(self, state):

        Q_value = self.Q_value.eval(feed_dict={self.state_input: [state]})[0]

        if self.epsilon > 0.1:
            self.epsilon = self.epsilon - 0.000005
        else:
            self.epsilon *= DECAY_RATE

        if random.random() <= self.epsilon:
            return random.randint(0, self.action_dim - 1)
        else:
            return np.argmax(Q_value)

    def action(self, state):
        if self.bool_defaule_action:
            return 2
        else:
            return np.argmax(self.Q_value.eval(feed_dict={self.state_input: [state]})[0])

    def store_data(self, state, action, reward, next_state, done):
        one_hot_action = np.zeros(self.action_dim)
        one_hot_action[action] = 1
        self.replay_buffer.append((state, one_hot_action, reward,
                                   next_state, done))
        if len(self.replay_buffer) > REPLAY_SIZE:
            self.replay_buffer.popleft()

